from logging import getLogger

import scipy.ndimage as scipy_ndimage

from mantidimaging import helper as h
from mantidimaging.core.parallel import shared_mem as psm
from mantidimaging.core.parallel import utility as pu
from mantidimaging.core.utility.progress_reporting import Progress


def _cli_register(parser):
    default_size = None
    size_help = "Apply median filter (2d) on \
                 reconstructed volume with the given kernel size."
    mode_help = "Default: %(default)s\n"\
        "Mode of median filter which determines how the array \
         borders are handled."

    parser.add_argument(
        "--median-size",
        type=int,
        required=False,
        default=default_size,
        help=size_help)

    parser.add_argument(
        "--median-mode",
        type=str,
        required=False,
        default=modes()[0],
        choices=modes(),
        help=mode_help)

    return parser


def modes():
    return ['reflect', 'constant', 'nearest', 'mirror', 'wrap']


def execute(data, size, mode=modes()[0], cores=None, chunksize=None,
            progress=None):
    """
    :param data: Input data as a 3D numpy.ndarray

    :param size: Size of the kernel

    :param mode: The mode with which to handle the endges.
                 One of [reflect, constant, nearest, mirror, wrap].

    :param cores: The number of cores that will be used to process the data.

    :param chunksize: The number of chunks that each worker will receive.

    :return: Returns the processed data

    """
    h.check_data_stack(data)

    if size and size > 1:
        if pu.multiprocessing_available():
            data = _execute_par(data, size, mode, cores, chunksize, progress)
        else:
            data = _execute_seq(data, size, mode, progress)

    h.check_data_stack(data)
    return data


def _execute_seq(data, size, mode, progress=None):
    log = getLogger(__name__)
    progress = Progress.ensure_instance(progress,
                                        num_steps=data.shape[0],
                                        task_name='Median filter')

    with progress:
        log.info("Median filter, with pixel data type: {0}, filter "
                 "size/width: {1}.".format(data.dtype, size))

        progress.add_estimated_steps(data.shape[0])
        for idx in range(0, data.shape[0]):
            data[idx] = scipy_ndimage.median_filter(data[idx], size, mode=mode)
            progress.update()

    return data


def _execute_par(data, size, mode, cores=None, chunksize=None, progress=None):
    log = getLogger(__name__)
    progress = Progress.ensure_instance(progress, task_name='Median filter')

    # create the partial function to forward the parameters
    f = psm.create_partial(
        scipy_ndimage.median_filter,
        fwd_func=psm.return_fwd_func,
        size=size,
        mode=mode)

    with progress:
        log.info("PARALLEL median filter, with pixel data type: {0}, filter "
                 "size/width: {1}.".format(data.dtype, size))

        progress.update()
        data = psm.execute(data, f, cores, chunksize, "Median Filter")

    return data
