from mantidimaging.core.utility.progress_reporting import Progress


def timed_import(config, progress=None):
    """
    Wraps the importing function in a timer.

    This function should not be used from the filters as it will add a lot of
    clutter to the stdout.

    :param config: A ReconstructionConfig with all the necessary parameters to
                   run a reconstruction.

    :return: the imported tool
    """
    progress = Progress.ensure_instance(progress, task_name='Tool Importing')

    with progress:
        progress.update(msg="Importing tool {0}".format(config.func.tool))
        tool = do_importing(config.func.tool, config.func.algorithm)

    return tool


def do_importing(tool, alg='fbp'):
    """
    The main importing function.

    Does the decision on which tool to import based on the input.

    :param tool: The name of the tool to be imported

    :param alg: The algorithm that will be used. This is only done the ensure
                that it is actually supported.
                If no algorithm is provided, the default is 'fbp', because it's
                supported by all tools.

    :return: the tool reference
    """
    if not isinstance(tool, str):
        raise TypeError(
            "The name of a reconstruction tool is required as a string. "
            "Got: {0}".format(tool))
    if 'tomopy' == tool:
        from mantidimaging.core.tools.tomopy_tool import TomoPyTool
        TomoPyTool.check_algorithm_compatibility(alg)
        imported_tool = TomoPyTool()

    elif 'astra' == tool:
        from mantidimaging.core.tools.astra_tool import AstraTool
        AstraTool.check_algorithm_compatibility(alg)
        imported_tool = AstraTool()
    else:
        raise ValueError("Tried to import unknown tool: {0}".format(tool))

    return imported_tool
